#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Channel

Producer/Consumer entities
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from functools import partial
from itertools import ifilter
from pika import BasicProperties

__all__ = ['Producer', 'Consumer']

# Project requirements
from sleipnir.core.utils import maybe_list, uuid
from sleipnir.core.async import engine, Task

# local submodule requirements
from .enums import ChannelTarget, ChannelTargetState, DeliveryMode
from .connections import BACKPRESSURE, BACKPRESSURE_MULTIPLIER
from .connections import ConnectionFactory
from .serialization import encode
from .compression import compress
from .messaging import Message


class ChannelEnd(object):

    def __init__(self, routes, connection=None):
        def __acquire_callback(connection):
            connection.channels.acquire(self._constructed)

        # fetch connection. Create a default one if required
        if not connection or type(connection) in (str, unicode):
            pool = ConnectionFactory.get_instance()
            conn = pool.create(connection) if connection else pool.default
            connection = conn

        # define watch containers, indexed by watcher id and also
        # indexed by group familiarity. This is required to be able to
        # implement the first (any) case
        self._watchers = []

        # now set variables
        self._routes = maybe_list(routes)

        # store connection temporarily into channel var so empty watch
        # calls works
        self._channel = connection

        # pending queue. Store here any early call to method when
        # channel isn't set properly yet
        self._pending = []

        # open a chanel if no one is available for this thread
        # wait till a channel is available for this connection
        if not connection.channels.default:
            connection.add_on_open_callback(__acquire_callback)

    def _constructed(self, channel):
        self._channel = channel
        # if watchers is empty, we assume that routes are up and
        # running so, we start to listen
        [pending() for pending in self._pending]

        if not self._watchers:
            self.when_resumed(None)

    def watch(self, builders=None, target=ChannelTarget.ROUTE,
              state=ChannelTargetState.RESUMED, first=True):

        pool = ConnectionFactory.get_instance().create(self._channel).builders
        wwid = pool.add_entity_watch(self, builders, target, state, first)
        self._watchers.append(wwid)
        return wwid

    def unwatch(self, wgid=[]):
        pool = ConnectionFactory.get_instance().create(self._channel).builders
        for wwid in maybe_list(wgid) or self._watchers:
            if wwid in self._watchers:
                pool.remove_watch(wwid)
                self._watchers.remove(wwid)

    @property
    def routes(self):
        return self._routes

    def when_declared(self, channel_end, frame):
        raise NotImplementedError

    def when_destroyed(self):
        raise NotImplementedError

    def when_resumed(self, route, frame):
        raise NotImplementedError

    def when_suspended(self):
        raise NotImplementedError


class Producer(ChannelEnd):

    def __init__(self, routes, connection=None, routing_key="",
                 serializer=None, compression=None,
                 on_return=None):

        # Call parent
        super(Producer, self).__init__(routes, connection)

        # init internal elements
        self._routing_key = ""
        self._serializer = serializer
        self._compression = compression
        self._on_return = on_return

    def _constructed(self, channel):
        # set on return callback
        channel.add_on_return_callback(self._on_return)
        # invoke parent method
        super(Producer, self)._constructed(channel)

    def _encode_payload(self, body, serializer=None, content_type=None,
                        content_encoding=None):

        # No content_type? Then we're serializing the data internally.
        if not content_type:
            serializer = serializer or self._serializer
            (content_type, content_encoding, body) = encode(body, serializer)
        else:
            # If the programmer doesn't want us to serialize,
            # make sure content_encoding is set.
            if isinstance(body, unicode):
                content_encoding = content_encoding or 'utf-8'
                body = body.encode(content_encoding)
            # If they passed in a string, we can't know anything
            # about it. So assume it's binary data.
            content_encoding = content_encoding or 'binary'

        # return formatted message
        return body, content_type, content_encoding

    def _compress_payload(self, body, compression=None, headers=None):
        if compression is None:
            # Allow compression = 'empty' string to force uncompress
            # when compress is enabled by default
            compression = self._compression
        if compression:
            body, headers["compression"] = compress(body, compression)
        return body

    def prepare(self, body, routing_key=None, mandatory=False, immediate=False,
                content_type=None, content_encoding=None, serializer=None,
                compression=None, headers=None,
                delivery_mode=DeliveryMode.PERSISTENT,
                **properties):

        # serialize payload
        args = (body, serializer, content_type, content_encoding)
        body, content_type, content_encoding = self._encode_payload(*args)

        # compress payload
        headers = headers or {}
        body = self._compress_payload(body, compression, headers)

        # prepare envelope
        exchange = self.exchange
        routing_key = routing_key or self._routing_key
        properties = BasicProperties(
            content_type=content_type,
            content_encoding=content_encoding,
            delivery_mode=int(delivery_mode),
            headers=headers,
            **properties)

        # return message
        return (exchange, routing_key, body, properties, mandatory, immediate,)

    @engine
    def delay_publish(self, message, callback, timeout=None):
        # wait required?
        connect = self._channel.slp_connection
        timeout = timeout if timeout is not None else connect.push_timeout
        if timeout > 0:
            yield Task(connect.add_timeout, timeout)
        #invoke
        self.publish(message)
        callback(message)
        # reset backpressure counter?
        if timeout:
            new_timeout = connect.push_timeout
            if timeout == connect.push_timeout:
                new_timeout /= BACKPRESSURE_MULTIPLIER
            if new_timeout <= BACKPRESSURE:
                new_timeout = 0
            connect.push_timeout = new_timeout

    def publish(self, message):
        if not hasattr(self._channel, "slp_connection"):
            pending = partial(self.publish, message)
            self._pending.append(pending)
            return

        if type(message) not in (list, tuple,):
            message = tuple(self.prepare(message))
        #publish message
        self._channel.basic_publish(*message)

    @property
    def exchange(self):
        return self._routes[0] if self._routes else ""

    @property
    def routing_key(self):
        return self._routing_key

    @property
    def serializer(self):
        return self._serializer

    @property
    def compression(self):
        return self._compression

    @property
    def on_return(self):
        return self._on_return

    @on_return.setter
    def on_return(self, value):
        if self._on_return:
            self._channel.remove_on_return_callback(self._on_return)
        self._on_return = value
        self._channel.add_on_return_callback(self._on_return)


class Consumer(ChannelEnd):
    def __init__(self, routes, connection=None, no_ack=True,
                 exclusive=False, callbacks=[], on_decode_error=None):

        # Call parent
        super(Consumer, self).__init__(routes, connection)

        # init internal elements
        self._no_ack = no_ack
        self._exclusive = exclusive
        self._callbacks = maybe_list(callbacks)
        self._on_decode_error = on_decode_error
        self._active_tags = {}

    def __iadd__(self, queues):
        self._routes += maybe_list(queues)
        raise NotImplementedError

    def __idel__(self, queues):
        for queue in maybe_list(queues):
            self._routes.remove(queue)

    def _receive(self, *args):
        try:
            message = Message(self._channel, args)
            decoded = message.payload
        except Exception, ex:
            if self._on_decode_error:
                self._on_decode_error(message, ex)
                return
            raise
        else:
            self.when_received(decoded, message)

    def append_callbacks(self, callbacks):
        self._callbacks += maybe_list(callbacks)

    def remove_callbacks(self, callbacks):
        for callback in maybe_list(callbacks):
            self._callbacks.remove(callback)

    def consume(self, no_ack=None, exclusive=None, timeout=None):
        if not hasattr(self._channel, "slp_connection"):
            pending = partial(self.consume, no_ack, exclusive, timeout)
            self._pending.append(pending)
            return

        no_ack = no_ack or self._no_ack
        exclus = exclusive or self._exclusive

        tags = []
        queue_filter = lambda x: x not in self._active_tags

        for queue in ifilter(queue_filter, self._routes):
            channel = self._channel
            tag = channel.basic_consume(self._receive, queue, no_ack, exclus)
            self._active_tags[queue] = tag
            tags.append(tag)

        if timeout:
            timeout_callback = partial(self.cancel, tags=tags)
            self._channel.slp_connection.add_timeout(timeout, self.cancel)

        # notify consume
        self.when_consumed(tags)
        return tags

    def cancel(self, queues=[], tags=[]):
        if tags:
            # lookup for valid tags
            for queue, tag in self._active_tags.iteritems():
                if tag in tags:
                    queues.append(queue)
        else:
            # no tags? use all active consumers or queues pased
            queues = queues or self._active_tags.keys()

        # cancel selected consumers
        for queue in queues:
            tag = self._active_tags.pop(queue)
            self._channel.basic_cancel(tag)
            tags.append(tag)

        # notify cancellation
        self.when_canceled(tags)

    def when_received(self, body, message):
        [callback(body, message) for callback in self._callbacks]

    def when_consumed(self, tags):
        pass

    def when_canceled(self, tags):
        pass

    @property
    def queues(self):
        return self._routes

    @property
    def consumer_tags(self):
        return self._active_tags.iteritems()



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Messaging

A wrapper around AMQP frames
"""

from __future__ import absolute_import

__author__  = "Ask Solem"
__license__ = "BSD like LICENSE"

# Import here any required modules.
from operator import attrgetter

# Inport local requirements
from .compression import decompress
from .serialization import decode


class MessageStateError(Exception):
    """The message has already been acknowledged."""
    pass


class AbstractMessage(object):
    """Base class for received messages."""

    ACKNOWLEDGED_STATES = frozenset([
            "ACK", "ACK_MULTIPLE", "REJECTED", "REQUEUED"
    ])

    _state = None

    MessageStateError = MessageStateError

    #: The channel the message was received on.
    channel = None

    #: Delivery tag used to identify the message in this channel.
    delivery_tag = None

    #: Content type used to identify the type of content.
    content_type = None

    #: Content encoding used to identify the text encoding of the body.
    content_encoding = None

    #: Additional delivery information.
    delivery_info = None

    #: Message headers
    headers = None

    #: Application properties
    properties = None

    #: Raw message body (may be serialized), see :attr:`payload` instead.
    body = None

    def __init__(self, channel, body=None, delivery_tag=None,
                 content_type=None, content_encoding=None, delivery_info={},
                 properties=None, headers=None, postencode=None,
                 **kwargs):

        self.channel = channel
        self.body = body
        self.delivery_tag = delivery_tag
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.delivery_info = delivery_info
        self.headers = headers or {}
        self.properties = properties or {}

        compression = self.headers.get("compression")
        if compression:
            self.body = decompress(self.body, compression)
        if postencode and isinstance(self.body, unicode):
            self.body = self.body.encode(postencode)

        self._decoded_cache = None
        self._state = "RECEIVED"

    def ack(self, multiple=False):
        """Acknowledge this message as being processed.,
        This will remove the message from the queue.

        :raises MessageStateError: If the message has already been
            acknowledged/requeued/rejected.

        """
        if self.acknowledged:
            msg = "Message already acknowledged with state: %s" % self._state
            raise self.MessageStateError(msg)
        # confirm
        if self.channel.slp_no_ack_consumers:
            no_ack_list = self.channel.slp_onsumers
            if self.delivery_info.get("consumer_tag") in no_ack_list:
                return
        # set ACK_MULTIPLE for all previous messages
        if multiple:
            self.channel.slp_last_ack_deliver = self.delivery_tag
        # send ACK
        self.channel.basic_ack(self.delivery_tag, multiple)
        self._state = "ACK"

    def reject(self):
        """
        Reject this message.

        The message will be discarded by the server.

        :raises MessageStateError: If the message has already been
            acknowledged/requeued/rejected.
        """
        if self.acknowledged:
            msg = "Message already acknowledged with state: %s" % self._state
            raise self.MessageStateError(msg)
        # reject message
        self.channel.basic_reject(self.delivery_tag, requeue=False)
        self._state = "REJECTED"

    def requeue(self):
        """
        Reject this message and put it back on the queue.

        You must not use this method as a means of selecting messages
        to process.

        :raises MessageStateError: If the message has already been
            acknowledged/requeued/rejected.
        """
        if self.acknowledged:
            msg = "Message already acknowledged with state: %s" % self._state
            raise self.MessageStateError(msg)
        # requeue message
        self.channel.basic_reject(self.delivery_tag, requeue=True)
        self._state = "REQUEUED"

    def decode(self):
        """
        Deserialize the message body, returning the original python
        structure sent by the publisher.
        """
        return decode(self.body, self.content_type, self.content_encoding)

    @property
    def acknowledged(self):
        """Set to true if the message has been acknowledged."""

        if self._state in self.ACKNOWLEDGED_STATES:
            return True

        last_deliver = self.channel.slp_last_ack_deliver
        if last_deliver and last_deliver > self.delivery_tag:
            self._state = "ACK_MULTIPLE"
            return True

        # no acknowledge
        return False

    @property
    def payload(self):
        """The decoded message body."""
        if not self._decoded_cache:
            self._decoded_cache = self.decode()
        return self._decoded_cache


class Message(AbstractMessage):

    PROPERTIES = (
        "app_id",
        "cluster_id",
        "content_encoding",
        "content_type",
        "correlation_id",
        "delivery_mode",
        "expiration",
        "headers",
        "message_id",
        "priority",
        "reply_to",
        "timestamp",
        "type",
        "user_id",
        )

    def __init__(self, channel, frame, **kwargs):
        channel_id, method, props, body = frame

        # prepare kwargs
        dinfo = dict(
            consumer_tag=getattr(method, "consumer_tag", None),
            routing_key=method.routing_key,
            delivery_tag=method.delivery_tag,
            redelivered=method.redelivered,
            exchange=method.exchange)
        attrs = dict(zip(self.PROPERTIES, attrgetter(*self.PROPERTIES)(props)))

        # update
        kwargs.update({
                "body":             body,
                "delivery_tag":     method.delivery_tag,
                "content_type":     props.content_type,
                "content_encoding": props.content_encoding,
                "headers":          props.headers,
                "properties":       attrs,
                "delivery_info":    dinfo,
                })

        super(Message, self).__init__(channel, **kwargs)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Connections

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from functools import partial

from pika import PlainCredentials, ConnectionParameters
from pika.adapters import SelectConnection

__all__ = ['ConnectionFactory']

# Project requirements
from sleipnir.core.factory import NamedAbstractFactory

# local submodule requirements
from .enums import ConnectionType, ConnectionState
from .pools import ChannelPool, BuilderPool
from .parser import connection_url, URI


BACKPRESSURE = 0.15
BACKPRESSURE_LIMIT = 90
BACKPRESSURE_MULTIPLIER = 3


class ConnectionError(Exception):
    """Connection exception"""


class ConnectionFactory(NamedAbstractFactory):
    """Connection factory"""

    _shared_backends = {}
    _shared_connections = {}

    def __init__(self):
        super(ConnectionFactory, self).__init__()
        self._last = []
        self._default = None

    def __getattribute__(self, name):
        if name == '_backends':
            name = '_shared_backends'
        if name == '_connections':
            name = '_shared_connections'
        return object.__getattribute__(self, name)

    def __getitem__(self, key):
        return self._connections.get(key, None)

    def __delitem__(self, key):
        if key in self._connections:
            self._connections[key].close()
        del self._connections[key]

    def _on_connected(self, conn):
        conn.state = ConnectionState.STARTED
        if conn.url in self._connections:
            # Set default conn if no one has been set
            self._default = self._default or conn
            self._last.append(conn)
            # invoke template method
            self.when_connected(conn)
            # add callbacks for close and backpressure events
            conn.add_on_close_callback(self._on_disconnected)
            conn.add_backpressure_callback(self._on_backpressure)

    def _on_disconnected(self, conn):
        conn.state = ConnectionState.CLOSED
        if conn.url in self._connections:
            self.when_disconnected(conn)
            self._last.remove(conn)

    def _on_backpressure(self, conn):
        if conn.url in self._connections:
            if not conn.push_timeout:
                conn.push_timeout = BACKPRESSURE
            conn.push_timeout *= BACKPRESSURE_MULTIPLIER
            conn.push_timeout = min(conn.push_timeout, BACKPRESSURE_LIMIT)
            # notify about pressure
            self.when_backpressure(conn)

    def _on_connection_timeout(self):
        for conn in self._connections.values():
            if conn.state == ConnectionState.STARTING:
                self.when_error(conn)
                del self._connections[conn.url]

    def when_connected(self, connection):
        pass

    def when_disconnected(self, connection):
        pass

    def when_backpressure(self, connection):
        pass

    def when_error(self, connection):
        pass

    @property
    def default(self):
        if not self._default:
            for conn in self._connections.itervalues():
                return conn
        return self._default

    @property
    def last(self):
        return self._last[-1] if self._last else None

    def create(self, url, conn_type=ConnectionType.ASYNC):
        # check that url pased (url_string, connection or channel)
        # doesn't exist in connections pool
        connection = url
        if hasattr(connection, 'slp_connection'):
            connection = url.slp_connection
        if isinstance(connection, AbstractConnection):
            url = connection.url

        # check now url
        if url not in self._connections:
            conn = super(ConnectionFactory, self).create(url, conn_type)
            self._connections[url] = conn
            # set callbacks
            conn.add_on_open_callback(self._on_connected)
            # add a timeout so if connection isn't valid, we should remove it
            conn.add_timeout(2, self._on_connection_timeout)
            conn.state = ConnectionState.STARTING
        return self._connections[url]


class MetaConnection(type):
    """Connection Metaclass"""

    def __init__(mcs, name, bases, dct):
        type.__init__(mcs, name, bases, dct)
        if not name.endswith("Connection"):
            ConnectionFactory.get_instance().register(name, mcs)


class AbstractConnection(object):
    """Prepares a connection. only get a pika connection is lazy created"""

    __metaclass__ = MetaConnection

    def __init__(self, url, limit=4):
        """Whenever a property changes, update stamp"""
        self._url = URI(url, connection_url)
        self._state = ConnectionState.CLOSED
        self._push_timeout = 0
        # set up pools
        self._builders_pool = BuilderPool(self)
        self._channels_pool = ChannelPool(self, limit)

    @property
    def type(self):
        return self.CONNECTION_TYPE

    @property
    def url(self):
        return str(self._url)

    @property
    def channels(self):
        return self._channels_pool

    @property
    def builders(self):
        return self._builders_pool

    @property
    def loop(self):
        # get connection variable to be used by start method
        connection = self

        def start(self, timeout=None):
            if timeout is not None:
                connection.add_timeout(timeout, connection.close)
            return self.__class__.start(self)

        # Start event loop
        if not hasattr(self.ioloop, "slp_monkey_patched"):
            self.ioloop.slp_monkey_patched = True
            klass = self.ioloop.__class__
            func_type = type(klass.start)
            self.ioloop.start = func_type(start, self.ioloop, klass)
        return self.ioloop

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def push_timeout(self):
        return self._push_timeout

    @push_timeout.setter
    def push_timeout(self, value):
        self._push_timeout = value

    def wait(self, state, callback):
        if state == self._state:
            return callback(self)
        if state == ConnectionState.STARTED:
            return self.add_on_open_callback(callback)
        if state == ConnectionState.CLOSED:
            return self.add_on_close_callback(callback)

        # Unsupported State
        raise NotImplementedError

    def routes(self, builder, **builder_kwargs):
        builder = builder(self, **builder_kwargs)
        return self._builders_pool.register(builder)

    @classmethod
    def can_handle(cls, url, conn_type):
        return cls.CONNECTION_TYPE == conn_type

    @classmethod
    def new(cls, url, conn_type):
        return cls(url)


class ConnectionMock(AbstractConnection):
    """A mock connection object"""

    CONNECTION_TYPE = ConnectionType.MOCK

    class IOLoopMock(object):
        def start(self):
            pass

    @property
    def loop(self):
        return IOLoopMock()


class PikaConnection(AbstractConnection):

    def __init__(self, url):
        AbstractConnection.__init__(self, url)

        credentials = None
        if self._url.username:
            credentials = PlainCredentials(
                self._url.username, self._url.password)

        for host, port in self._url.hosts():
            self._parameters = ConnectionParameters(
                credentials=credentials,
                host=host,
                port=port,
                virtual_host=self._url.vhost)

            #Fixme: Support multiple hosts
            break


class ConnectionAsync(PikaConnection, SelectConnection):

    CONNECTION_TYPE = ConnectionType.ASYNC

    def __init__(self, url):
        PikaConnection.__init__(self, url)
        SelectConnection.__init__(self, self._parameters)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
amqparser

A handy way to parse amqp urls
"""

from __future__ import absolute_import

__author__  = "Ask Solem"
__license__ = "BSD, See LICENSE file for details"

# Import here any required modules.
try:
    from urlparse import urlparse, urlsplit
    from urlparse import parse_qsl
except ImportError:
    from cgi import parse_qsl  # noqa

# Project requirements
from sleipnir.core.utils import kwdict, uuid

__all__ = ['amqparse', 'routeparse']


def binding_url(url):
    """
    https://cwiki.apache.org/qpid/bindingurlformat.html
    <exchange_class>://[<exchange_name>][/QueueName][?options]
    """
    parts = urlsplit(url)
    return dict({"name": uuid(),
                 "type": parts.scheme,
                 "origin": parts.netloc or "",
                 "destination": parts.path[1:] if parts.path else ""},
                **kwdict(dict(parse_qsl(parts.query))))


def connection_url(url):
    """
    amqp_url       = "amqp://" [ userinfo "@" ] addr_list [ vhost ]
    addr_list      = addr *( "," addr )
    addr           = prot_addr [ options ]
    prot_addr      = tcp_prot_addr | other_prot_addr
    vhost          = "/" *pchar [ options ]

    tcp_prot_addr  = tcp_id tcp_addr
    tcp_id         = "tcp:" / ""	; tcp is the default
    tcp_addr       = [ host [ ":" port ] ]

    other_prot_addr= other_prot_id ":" *pchar
    other_prot_id  = scheme

    options        = "?" option *( ";" option )
    option         = name "=" value
    name           = *pchar
    value          = *pchar
    """

    parts = urlsplit(url)
    addrlist = parts.netloc
    password = user = ""

    if '@' in addrlist:
        auth, _, addrlist = addrlist.partition('@')
        user, _, password = auth.partition(':')

    return dict({"type": parts.scheme,
                 "user": user,
                 "password": password,
                 "hosts": addrlist.split(','),
                 "vhost": parts.path if parts.path else ""},
                **kwdict(dict(parse_qsl(parts.query))))


class URI(object):
    def __init__(self, url, parser):
        self._urlstr = url
        self._urldct = parser(self._urlstr)

    def __str__(self):
        return self._urlstr

    def __contains__(self, item):
        return item in self._urldct

    def __iter__(self):
        return self._urldct.iteritems()

    def __getattr__(self, key):
        return self._urldct.get(key, '')

    def hosts(self):
        for addr in self._urldct["hosts"]:
            host, _, port = addr.partition(':')
            yield host, int(port or 5672)


# sugar method to register handlers
def register_scheme(scheme):
    import urlparse
    for method in filter(lambda s: s.startswith('uses_'), dir(urlparse)):
        getattr(urlparse, method).insert(0, scheme)

#connection schemas
register_scheme('amqp')
register_scheme('amqp-dev')

# binding url schemas
register_scheme('direct')
register_scheme('direct-exchange')
register_scheme('fanout')
register_scheme('fanout-exchange')
register_scheme('topic')
register_scheme('topic-exchange')



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
builder

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from functools import partial
from itertools import ifilter

__all__ = ['Builder']

# Project requirements
from sleipnir.core.factory import NamedAbstractFactory
from sleipnir.core.utils import maybe_list, uuid
from sleipnir.core.async import Task, engine

# local submodule requirements
from .enums import ChannelTargetState, ChannelTarget, RouteState
from .enums import RouteEndType, RouteType, RouteState
from .routes import Route


class BrokerFactory(NamedAbstractFactory):
    """Broker private"""


class MetaBroker(type):
    """Broker Metaclass"""
    def __init__(mcs, name, bases, dct):
        type.__init__(mcs, name, bases, dct)
        if name.endswith("Broker"):
            mcs.__merge(mcs.mro())
            BrokerFactory.get_instance().register(name, mcs)

    def __merge(mcs, mro):
        if 'attrs' in dir(mcs):
            mro_filter = lambda x: mcs != x and 'attrs' in dir(x)
            for pcs in ifilter(mro_filter, mro):
                new_attrs = pcs.attrs.copy()
                for key, value in mcs.attrs.iteritems():
                    new_attrs.setdefault(key, {}).update(value)
                mcs.attrs = new_attrs


class BrokerAbstract(object):

    __metaclass__ = MetaBroker

    attrs = {
        RouteEndType.EXCHANGE: {
            'type':               ('type',        None, None),
            'passive':            ('passive',     bool, None),
            'durable':            ('durable',     bool, None),
            'auto-delete':        ('auto-delete', bool, None),
            'internal':           ('internal',    bool, None),
            'no-wait':            ('no-wait',     bool, None),
            },
        RouteEndType.QUEUE: {
            'passive':            ('passive',     bool, None),
            'durable':            ('durable',     bool, None),
            'auto-delete':        ('auto-delete', bool, None),
            'exclusive':          ('exclusive',   bool, None),
            'no-wait':            ('no-wait',     bool, None),
            },
        RouteType.QUEUE: {
            'origin':             ('exchange',    None, None),
            'destination':        ('queue',       None, None),
            'no-wait':            ('no-wait',     None, None),
            'routing-key':        ('routing-key', None, None),
            }
        }

    def create_entity(self, entity, callback=None):
        raise NotImplementedError

    def delete_entity(self, entity, callback=None):
        raise NotImplementedError

    def resume_route(self, route, callback=None):
        raise NotImplementedError

    def suspend_route(self, route, callback=None):
        raise NotImplementedError


class PikaScaffoldBroker(BrokerAbstract):
    """Simple broker"""

    def _type(value):
        extension = "-exchange"
        return value[:len(extension)] if value.endswith(extension) else value

    attrs = {
        RouteEndType.EXCHANGE: {
            'destination':        ('exchange',    None,  None),
            'origin':             ('exchange',    None,  None),
            'type':               ('type',        _type, False),
            'auto-delete':        ('auto_delete', bool,  None),
            'no-wait':            ('nowait',      bool,  None),
            # extra parameters
            'alternate-exchange': ('alternate-exchange', None, 'params')
            },
        RouteEndType.QUEUE: {
            'destination':        ('queue',       None,  None),
            'auto-delete':        ('auto_delete', bool,  None),
            'no-wait':            ('nowait',      bool,  None),
            },
        RouteType.QUEUE: {
            'no-wait':            ('nowait',      bool,  None),
            'routing-key':        ('routing_key', None,  None),
            }
        }

    @classmethod
    def can_handle(cls, connection, **kwargs):
        url = connection.url
        return url.startswith('amqp:') and not kwargs.get("force", False)


class PikaBroker(PikaScaffoldBroker):
    """defines primitves to build routes"""

    def __init__(self, connection, **kwargs):
        self._connection = connection

    @engine
    def create_entity(self, entity, callbacks):
        # peek thread channel
        channel = self._connection.channels.default
        # declare entities
        if entity.end_type == RouteEndType.EXCHANGE:
            frame = yield Task(channel.exchange_declare, **entity.params)
            [callback(entity, frame) for callback in callbacks]
        if entity.end_type == RouteEndType.QUEUE:
            frame = yield Task(channel.queue_declare, **entity.params)
            [callback(entity, frame) for callback in callbacks]

    @engine
    def resume_route(self, route, callbacks):
        # peek thread channel
        channel = self._connection.channels.default
        rparams = route.params(self.attrs)
        # bind routes
        if route.route_type == RouteType.QUEUE:
            frame = yield Task(channel.queue_bind, **rparams)
            [callback(route, frame) for callback in callbacks]
        if route.route_type == RouteType.EXCHANGE:
            frame = yield Task(channel.exchange_bind, **rparams)
            [callback(route, frame) for callback in callbacks]

    @classmethod
    def can_handle(cls, connection, **kwargs):
        url = connection.url
        return url.startswith('amqp:')    \
            if kwargs.get("force", False) \
            else url.startswith('amqp-dev:')


class MockBroker(BrokerAbstract):
    @classmethod
    def can_handle(cls, url, **kwargs):
        url = connection.url
        return url.startswith('mock:')


class Builder(object):

    class RouteBuilder(object):

        def __init__(self, route, builder):
            self._route = route if isinstance(route, Route) else Route(route)
            self._build = builder

        def create(self, callback=None):
            broker = self._build._broker
            brokcb = maybe_list(callback)
            brokcb.append(self._build.when_created)

            # don't declare again if already defined
            if self._route.state in (RouteState.BIND, RouteState.UNBIND):
                [callback(self._route, None) for callback in brokcb]
                return

            # append watchers
            watchers = self._build._watchers[ChannelTargetState.DECLARED]
            brokcb += watchers.itervalues()

            # process
            if self._route.state != RouteState.ORIGIN_DECLARED:
                entity = self._route(broker.attrs).origin
                broker.create_entity(entity, brokcb)

            if self._route.state != RouteState.DESTINATION_DECLARED:
                entity = self._route(broker.attrs).destination
                broker.create_entity(entity, brokcb)

            # update state
            self._route.state = RouteState.UNBIND

        @engine
        def resume(self, callback=None):
            broker = self._build._broker
            brokcb = maybe_list(callback)
            brokcb.append(self._build.when_resumed)

            # append watchers
            watchers = self._build._watchers[ChannelTargetState.RESUMED]
            brokcb += watchers.itervalues()

            # don't resume again if already defined
            if self._route.state in (RouteState.BIND,):
                [callback(self._route) for callback in brokcb]
                return

            # we assume that we are on unbind state
            broker.resume_route(self._route, brokcb)

        def destroy(self, callback=None):
            raise NotImplementedError

    @engine
    def __init__(self, connection, **broker_kwargs):
        # create broker and init properties
        factory = BrokerFactory.get_instance()
        self._broker = factory.create(connection, **broker_kwargs)
        self._routes = []
        self._watchers = {
            ChannelTargetState.DECLARED:  {},
            ChannelTargetState.RESUMED:   {},
            ChannelTargetState.SUSPENDED: {},
            ChannelTargetState.DESTROYED: {},
            }

        # Channel is still undefined
        self._connection = connection

        # open a chanel if no one is available for this thread
        channel_pool = self._connection.channels
        if channel_pool.default is None:
            yield Task(self._connection.add_on_open_callback)
            yield Task(channel_pool.acquire)

        # build routes
        yield Task(self.create)
        yield Task(self.resume)

    def __iter__(self):
        return iter(self._routes)

    # virtual template methods
    def when_created(self, route, frame):
        pass

    def when_resumed(self, route, frame):
        route.state = RouteState.BIND

    def when_destroyed(self, route):
        pass

    def __call__(self, route):
        route = self.RouteBuilder(route, self)
        self._routes.append(route)
        return self

    # routes status
    def create(self, callback=None):
        """Bind all routes"""
        channel = self._connection.channels.default
        [route.create(callback) for route in self._routes]

    def resume(self, callback=None):
        channel = self._connection.channels.default
        [route.resume(callback) for route in self._routes]

    def destroy(self, callback=None):
        """Destroy route elements"""

    # watchers stuff
    def add_entity_watch(self, channel_end, target, state):
        def mentity(channel_end, route, frame):
            if hasattr(channel_end, 'exchange'):
                if channel_end.exchange == route.name:
                    channel_end.when_declared(route, frame)
            else:
                raise NotImplementedError

        def mroute(channel_end, route, frame):
            if hasattr(channel_end, 'exchange'):
                if route.origin.name in channel_end.exchange:
                    channel_end.when_resumed(route, frame)
            elif hasattr(channel_end, 'queues'):
                if route.destination.name in channel_end.queues:
                    channel_end.when_resumed(route, frame)
            else:
                raise NotImplementedError

        wid = uuid()
        matchfun = mentity if target == ChannelTarget.ENTITY else mroute
        self._watchers[state][wid] = partial(matchfun, channel_end)
        return wid



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""Sleipnir Transport constants"""

from __future__ import absolute_import

__author__           = 'Carlos Martín <cmartin@liberalia.net>'
__version__          = '0.0.90'
__date__             = '2011-10-05'
__license__          = 'GNU General Public License, version 2'

__namespace__        = "sleipnir"
__modname__          = "transport"
__appname__          = __namespace__ + '.' + __modname__
__title__            = 'Sleipnir Transport'
__release__          = '1'
__summary__          = 'A Wrapper around pika AMQP Library'
__url__              = 'http://sleipnir.liberalia.net/'
__copyright__        = '© 2011 Carlos Martín'

__classifiers__      = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Topic :: Software Development',
    ]

__long_description__ = """\
Add Here a a description to this package
"""

__requires__ = [
    'pika                     >= 0.9.6-pre0',
    'anyjson                  >= 0.3.0',
    __namespace__ + '.core    >= 0.1.92',
    ]

__tests_requires__ = [
    'pika                     >= 0.9.6-pre0',
    'anyjson                  >= 0.3.0',
    __namespace__ + '.core    >= 0.1.92',
    __namespace__ + '.testing >= 0.1rc6',
    ]



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Pools

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from Queue import Queue, Empty
from thread import get_ident
from threading import Lock

__all__ = ['BuilderPool', 'ChannelPool']

# Project requirements
from sleipnir.core.utils import uuid, maybe_list


class BuilderPool(object):
    def __init__(self, connection):
        self.__connection = connection
        self._builders = {}
        self._entities = {}
        self._watchers = {}

    def register(self, builder):
        # set default values
        self._builders.setdefault(get_ident(), []).append(builder)
        self._entities.setdefault(get_ident(), {})
        # add watch
        for entity, values in self._entities[get_ident()].iteritems():
            self.add_entity_watch(entity, builder, **values)
        return builder

    def add_entity_watch(self, entity, builders, target, state, first):
        wbls, wgid = [], uuid()

        # get a list/tuple of builders
        pool = maybe_list(builders) or self._builders.get(get_ident(), [])
        if type(pool) not in (list, tuple,):
            pool = [pool]
        # create watchers
        for builder in pool:
            wid = builder.add_entity_watch(entity, target, state)
            self._watchers.setdefault(wgid, wbls).append((wid, builder,))
        # store params to listen to new declared builders
        if builders is None:
            kwargs = {'target': target, 'state': state, 'first': first}
            self._entities.setdefault(get_ident(), {})
            self._entities[get_ident()][entity] = kwargs
        return wgid

    def __iter__(self):
        return iter(self._builders.get(get_ident(), []))


class ChannelPool(object):
    def __init__(self, connection, limit):
        self._limit = limit
        self._cconn = connection
        self._queue = Queue(self._limit)
        self._qlock = Lock()
        self._cpool = {}

    def __iter__(self):
        return self._cpool.itervalues()

    def __getattr__(self, name):
        getattr(self._cpool, name)

    def _consume_channel(self):
        # block until a channel is available
        try:
            for thread, callbacks in self._cpool.iteritems():
                if type(callbacks) in (list, tuple):
                    channel = self._queue.get(False)
                    self._cpool[thread] = channel
                    [callback(channel) for callback in callbacks]
                    return
            else:
                raise ConnectionError("Orphan channel")
        except Empty:
            pass

    def _produce_channel(self, channel):
        # Add custom properties
        channel.slp_connection = self._cconn
        channel.slp_last_ack_deliver = None
        channel.slp_no_ack_consumers = []

        self._queue.put(channel)
        self._consume_channel()

    def _release_channel(self):
        rthread = get_ident()
        channel = self._cpool[rthread]
        assert channel and type(channel) not in (list, tuple)
        del self._cpool[rthread]
        self._queue.put(channel)
        self._queue.task_done()
        self._consume_channel()

    @property
    def default(self):
        retval = self._cpool.get(get_ident(), None)
        return retval if type(retval) not in (list, tuple) else None

    def acquire(self, callback):
        # shortcut to get assigned channel
        thread = get_ident()
        channel = self._cpool.setdefault(thread, [])
        if type(channel) not in (list, tuple):
            callback(channel)
            return

        # register callback
        assert callback not in self._cpool
        self._cpool[thread].append(callback)

        # produce channel if required
        self._qlock.acquire()
        if (len(self._cpool) - 1) + self._queue.qsize() < self._limit:
            if len(self._cpool[thread]) == 1:
                self._cconn.channel(self._produce_channel)
        self._qlock.release()

    def release(self):
        self._release_channel(get_ident())

    def close(self):
        raise NotImplementedError
